r"""Contain the implementation of a section to analyze a column with
continuous values."""

from __future__ import annotations

__all__ = ["ColumnContinuousAdvancedSection"]

import logging
from typing import TYPE_CHECKING

import numpy as np
from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.section.base import BaseSection
from flamme.section.continuous import (
    add_cdf_plot,
    create_boxplot_figure,
    create_histogram_figure,
    create_stats_table,
)
from flamme.section.utils import (
    GO_TO_TOP,
    auto_yscale_continuous,
    compute_statistics,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels
from flamme.utils.range import find_range

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd


logger = logging.getLogger(__name__)


class ColumnContinuousAdvancedSection(BaseSection):
    r"""Implement a section that analyzes a continuous distribution of
    values.

    Args:
        series: The series/column to analyze.
        column: The column name.
        nbins: The number of bins in the histogram.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.section import ColumnContinuousSection
    >>> section = ColumnContinuousAdvancedSection(
    ...     series=pd.Series([np.nan, *list(range(101)), np.nan]), column="col"
    ... )
    >>> section
    ColumnContinuousAdvancedSection(
      (column): col
      (nbins): None
      (yscale): auto
      (figsize): None
    )
    >>> section.get_statistics()
    {'count': 103, 'num_nulls': 2, 'nunique': 102, 'mean': 50.0, 'std': 29.30...,
     'skewness': 0.0, 'kurtosis': -1.200235294117647, 'min': 0.0,
     'q001': 0.1, 'q01': 1.0, 'q05': 5.0, 'q10': 10.0, 'q25': 25.0, 'median': 50.0,
     'q75': 75.0, 'q90': 90.0, 'q95': 95.0, 'q99': 99.0, 'q999': 99.9, 'max': 100.0,
     '>0': 100, '<0': 0, '=0': 1, 'num_non_nulls': 101}

    ```
    """

    def __init__(
        self,
        series: pd.Series,
        column: str,
        nbins: int | None = None,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._series = series
        self._column = column
        self._nbins = nbins
        self._yscale = yscale
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "column": self._column,
                    "nbins": self._nbins,
                    "yscale": self._yscale,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def column(self) -> str:
        return self._column

    @property
    def yscale(self) -> str:
        return self._yscale

    @property
    def nbins(self) -> int | None:
        return self._nbins

    @property
    def series(self) -> pd.Series:
        return self._series

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict[str, float]:
        return compute_statistics(self._series)

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(f"Rendering the continuous distribution of {self._column}")
        stats = self.get_statistics()
        null_values_pct = (
            f"{100 * stats['num_nulls'] / stats['count']:.2f}" if stats["count"] > 0 else "N/A"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "table": create_stats_table(stats=stats, column=self._column),
                "total_values": f"{stats['count']:,}",
                "unique_values": f"{stats['nunique']:,}",
                "null_values": f"{stats['num_nulls']:,}",
                "null_values_pct": null_values_pct,
                "full_histogram": self._create_full_histogram(stats),
                "iqr_histogram": self._create_iqr_histogram(),
                "full_boxplot": self._create_full_boxplot(),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the discrete distribution of values for column <em>{{column}}</em>.

<ul>
  <li> total values: {{total_values}} </li>
  <li> number of unique values: {{unique_values}} </li>
  <li> number of null values: {{null_values}} / {{total_values}} ({{null_values_pct}}%) </li>
</ul>

<p style="margin-top: 1rem;">
<b> Analysis of the distribution </b>

{{full_histogram}}
{{full_boxplot}}
{{table}}
<p style="margin-top: 1rem;">

<p style="margin-top: 1rem;">
<b> Analysis of distribution in the inter-quartile range (IQR) </b>

{{iqr_histogram}}

"""

    def _create_full_boxplot(self) -> str:
        return create_boxplot_figure(
            series=self._series,
            xmin="q0",
            xmax="q1",
            figsize=self._figsize,
        )

    def _create_full_histogram(self, stats: dict) -> str:
        return create_histogram_figure(
            series=self._series,
            column=self._column,
            stats=stats,
            nbins=self._nbins,
            xmin="q0",
            xmax="q1",
            yscale=self._yscale,
            figsize=self._figsize,
        )

    def _create_iqr_histogram(self) -> str:
        return create_histogram_range_figure(
            series=self._series,
            column=self._column,
            nbins=self._nbins,
            xmin="q0.25",
            xmax="q0.75",
            yscale=self._yscale,
            figsize=self._figsize,
        )


def create_histogram_range_figure(
    series: pd.Series,
    column: str,
    nbins: int | None = None,
    yscale: str = "auto",
    xmin: float | str | None = None,
    xmax: float | str | None = None,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Create the HTML code of a histogram figure.

    Args:
        series: The series/column to analyze.
        column: The column name.
        nbins: The number of bins in the histogram.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The HTML code of the figure.
    """
    array = series.to_numpy()
    if array.size == 0:
        return "<span>&#9888;</span> No figure is generated because the column is empty"
    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    array_range = array[np.logical_and(array >= xmin, array <= xmax)]
    fig, ax = plt.subplots(figsize=figsize)
    ax.hist(array_range, bins=nbins, range=[xmin, xmax], color="tab:blue", alpha=0.9)
    if xmin < xmax:
        ax.set_xlim(xmin, xmax)
    readable_xticklabels(ax, max_num_xticks=100)
    ax.set_title(f"Distribution of values for column {column}")
    ax.set_ylabel("Number of occurrences")
    if yscale == "auto":
        yscale = auto_yscale_continuous(array=array_range, nbins=nbins)
    ax.set_yscale(yscale)
    add_cdf_plot(ax, array=array, nbins=nbins, xmin=xmin, xmax=xmax)
    return figure2html(fig, close_fig=True)



r"""Contain the implementation of a section to analyze the duplicate
values."""

from __future__ import annotations

__all__ = ["DuplicatedRowSection"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd


logger = logging.getLogger(__name__)


class DuplicatedRowSection(BaseSection):
    r"""Implement a section to analyze the number of duplicated rows.

    Args:
        frame: The DataFrame to analyze.
        columns: The columns used to compute the duplicated rows.
            ``None`` means all the columns.
        figsize: The figure
            size in inches. The first dimension is the width and the
            second is the height.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> import numpy as np
    >>> from flamme.section import DuplicatedRowSection
    >>> section = DuplicatedRowSection(
    ...     frame=pd.DataFrame(
    ...         {
    ...             "col1": np.array([1.2, 4.2, 4.2, 2.2]),
    ...             "col2": np.array([1, 1, 1, 1]),
    ...             "col3": np.array([1, 2, 2, 2]),
    ...         }
    ...     )
    ... )
    >>> section
    DuplicatedRowSection(
      (columns): None
      (figsize): None
    )
    >>> section.get_statistics()
    {'num_rows': 4, 'num_unique_rows': 3}

    ```
    """

    def __init__(
        self,
        frame: pd.DataFrame,
        columns: Sequence[str] | None = None,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._frame = frame
        self._columns = columns if columns is None else tuple(columns)
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping({"columns": self._columns, "figsize": self._figsize}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def frame(self) -> pd.DataFrame:
        r"""The DataFrame to analyze."""
        return self._frame

    @property
    def columns(self) -> tuple[str, ...] | None:
        r"""Tuple or ``None``: The columns used to compute the
        duplicated rows."""
        return self._columns

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        frame_no_duplicate = self._frame.drop_duplicates(subset=self._columns)
        return {"num_rows": self._frame.shape[0], "num_unique_rows": frame_no_duplicate.shape[0]}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(f"Rendering the duplicated rows section using the columns: {self._columns}")
        stats = self.get_statistics()
        columns = self._frame.columns if self._columns is None else self._columns
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "columns": ", ".join(columns),
                "num_columns": f"{len(columns):,}",
                "table": create_duplicate_table(
                    num_rows=stats["num_rows"], num_unique_rows=stats["num_unique_rows"]
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section shows the number of unique and duplicated rows when considering the following
{{num_columns}} columns: <em>{{columns}}</em>.

{{table}}

<p style="margin-top: 1rem;">
"""


def create_duplicate_histogram(
    num_rows: int, num_unique_rows: int, figsize: tuple[float, float] | None = None
) -> str:
    fig, ax = plt.subplots(figsize=figsize)
    x = list(range(3))
    ax.bar(x, [num_rows, num_unique_rows, num_rows - num_unique_rows], color="tab:blue")
    ax.set_xticks(x, labels=["total", "unique", "duplicate"])
    ax.set_xlim(-0.5, 2.5)
    ax.set_ylabel("Number of occurrences")
    ax.set_title("Analysis of the number of duplicate rows")
    return figure2html(fig, close_fig=True)


def create_duplicate_table(num_rows: int, num_unique_rows: int) -> str:
    r"""Create a table with information about duplicated rows.

    Args:
        num_rows (int): The number of rows.
        num_unique_rows (int): The number of unique rows.

    Returns:
        The HTML representation of the table.
    """
    num_duplicated_rows = num_rows - num_unique_rows
    pct_unique_rows = num_unique_rows / num_rows if num_rows else float("nan")
    pct_duplicated_rows = 1.0 - pct_unique_rows
    return Template(
        """
<table class="table table-hover table-responsive w-auto" >
<thead class="thead table-group-divider">
    <tr>
        <th>number of rows</th>
        <th>number of unique rows</th>
        <th>number of duplicated rows</th>
    </tr>
</thead>
<tbody class="tbody table-group-divider">
    <tr>
        <td {{num_style}}>{{num_rows}}</td>
        <td {{num_style}}>{{num_unique_rows}} ({{pct_unique_rows}}%)</td>
        <td {{num_style}}>{{num_duplicated_rows}} ({{pct_duplicated_rows}}%)</td>
    </tr>
    <tr class="table-group-divider"></tr>
</tbody>
</table>
"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "num_rows": f"{num_rows:,}",
            "num_unique_rows": f"{num_unique_rows:,}",
            "num_duplicated_rows": f"{num_duplicated_rows:,}",
            "pct_unique_rows": f"{100 * pct_unique_rows:.2f}",
            "pct_duplicated_rows": f"{100 * pct_duplicated_rows:.2f}",
        }
    )



r"""Implement some utility functions to manage optional dependencies."""

from __future__ import annotations

__all__ = [
    "check_clickhouse_connect",
    "is_clickhouse_connect_available",
    "clickhouse_connect_available",
]

from importlib.util import find_spec
from typing import TYPE_CHECKING, Any

from coola.utils.imports import decorator_package_available

if TYPE_CHECKING:
    from collections.abc import Callable


##############################
#     clickhouse_connect     #
##############################


def is_clickhouse_connect_available() -> bool:
    r"""Indicate if the ``clickhouse_connect`` package is installed or
    not.

    Returns:
        ``True`` if ``clickhouse_connect`` is available otherwise
            ``False``.

    Example usage:

    ```pycon
    >>> from flamme.utils.imports import is_clickhouse_connect_available
    >>> is_clickhouse_connect_available()

    ```
    """
    return find_spec("clickhouse_connect") is not None


def check_clickhouse_connect() -> None:
    r"""Check if the ``clickhouse_connect`` package is installed.

    Raises:
        RuntimeError: if the ``clickhouse_connect`` package is not
            installed.

    Example usage:

    ```pycon
    >>> from flamme.utils.imports import check_clickhouse_connect
    >>> check_clickhouse_connect()

    ```
    """
    if not is_clickhouse_connect_available():
        msg = (
            "`clickhouse_connect` package is required but not installed. "
            "You can install `clickhouse_connect` package with the command:\n\n"
            "pip install clickhouse-connect\n"
        )
        raise RuntimeError(msg)


def clickhouse_connect_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if
    ``clickhouse_connect`` package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``clickhouse_connect`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon
    >>> from flamme.utils.imports import clickhouse_connect_available
    >>> @clickhouse_connect_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_clickhouse_connect_available)



r"""Contain ``pandas.Series`` transformers to transform numeric
columns."""

from __future__ import annotations

__all__ = ["ToNumericSeriesTransformer"]

from typing import Any

import pandas as pd

from flamme.transformer.series.base import BaseSeriesTransformer


class ToNumericSeriesTransformer(BaseSeriesTransformer):
    r"""Implement a ``pandas.Series`` transformer to convert a
    ``pandas.Series`` to numeric type.

    Args:
        **kwargs: The keyword arguments for
            ``pandas.to_numeric``.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.transformer.series import ToNumeric
    >>> transformer = ToNumeric()
    >>> transformer
    ToNumericSeriesTransformer()
    >>> series = pd.Series(["1", "2", "3", "4", "5"])
    >>> series.dtype
    dtype('O')
    >>> out = transformer.transform(series)
    >>> out.dtype
    dtype('int64')

    ```
    """

    def __init__(self, **kwargs: Any) -> None:
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = ", ".join([f"{key}={value}" for key, value in self._kwargs.items()])
        return f"{self.__class__.__qualname__}({args})"

    def transform(self, series: pd.Series) -> pd.Series:
        return pd.to_numeric(series, **self._kwargs)



r"""Implement an analyzer that generates a section to analyze the data
types of each column."""

from __future__ import annotations

__all__ = ["DataTypeAnalyzer"]

import logging
from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import DataTypeSection
from flamme.utils.dtype import frame_column_types

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class DataTypeAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to find all the value types in each column.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import DataTypeAnalyzer
    >>> analyzer = DataTypeAnalyzer()
    >>> analyzer
    DataTypeAnalyzer()
    >>> frame = pd.DataFrame(
    ...     {
    ...         "int": np.array([np.nan, 1, 0, 1]),
    ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...         "str": np.array(["A", "B", None, np.nan]),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def analyze(self, frame: pd.DataFrame) -> DataTypeSection:
        logger.info("Analyzing the data types...")
        return DataTypeSection(dtypes=frame.dtypes.to_dict(), types=frame_column_types(frame))



r"""Contain the implementation of a section to analyze the temporal
distribution of a column with continuous values."""

from __future__ import annotations

__all__ = ["ColumnTemporalContinuousSection"]

import logging
from typing import TYPE_CHECKING

import numpy as np
from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    auto_yscale_continuous,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels
from flamme.utils.mathnan import remove_nan

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd


logger = logging.getLogger(__name__)


class ColumnTemporalContinuousSection(BaseSection):
    r"""Implement a section that analyzes the temporal distribution of a
    column with continuous values.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.section import ColumnContinuousSection
    >>> section = ColumnTemporalContinuousSection(
    ...     frame=pd.DataFrame(
    ...         {
    ...             "col": np.array([1.2, 4.2, np.nan, 2.2]),
    ...             "datetime": pd.to_datetime(
    ...                 ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
    ...             ),
    ...         }
    ...     ),
    ...     column="col",
    ...     dt_column="datetime",
    ...     period="M",
    ... )
    >>> section
    ColumnTemporalContinuousSection(
      (column): col
      (dt_column): datetime
      (period): M
      (yscale): auto
      (figsize): None
    )
    >>> section.get_statistics()
    {}

    ```
    """

    def __init__(
        self,
        frame: pd.DataFrame,
        column: str,
        dt_column: str,
        period: str,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._frame = frame
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._yscale = yscale
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "column": self._column,
                    "dt_column": self._dt_column,
                    "period": self._period,
                    "yscale": self._yscale,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def column(self) -> str:
        return self._column

    @property
    def dt_column(self) -> str:
        return self._dt_column

    @property
    def yscale(self) -> str:
        return self._yscale

    @property
    def period(self) -> str:
        return self._period

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            f"Rendering the temporal continuous distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "dt_column": self._dt_column,
                "period": self._period,
                "figure": create_temporal_figure(
                    frame=self._frame,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                    yscale=self._yscale,
                    figsize=self._figsize,
                ),
                "table": create_temporal_table(
                    frame=self._frame,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of column <em>{{column}}</em>
by using the column <em>{{dt_column}}</em>.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_figure(
    frame: pd.DataFrame,
    column: str,
    dt_column: str,
    period: str,
    yscale: str = "auto",
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Create a HTML representation of a figure with the temporal value
    distribution.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The HTML representation of the figure.
    """
    if frame.shape[0] == 0:
        return "<span>&#9888;</span> No figure is generated because the column is empty"
    array = frame[column].to_numpy(dtype=float)
    frame = frame[[column, dt_column]].copy()
    frame[dt_column] = frame[dt_column].dt.to_period(period).astype(str)
    frame_group = (
        frame.groupby(dt_column)[column]
        .apply(list)
        .reset_index(name=column)
        .sort_values(by=[dt_column])
    )

    data = [remove_nan(x) for x in frame_group[column].tolist()]
    labels = frame_group[dt_column].tolist()
    fig, ax = plt.subplots(figsize=figsize)
    ax.boxplot(
        data,
        notch=True,
        vert=True,
        widths=0.7,
        patch_artist=True,
        boxprops={"facecolor": "lightblue"},
    )
    ax.set_ylim(np.nanmin(array), np.nanmax(array))
    ax.set_xticks(np.arange(len(labels)), labels=labels)
    ax.set_title(f"Distribution of values for column {column}")
    if yscale == "auto":
        yscale = auto_yscale_continuous(array=array, nbins=100)
    ax.set_yscale(yscale)
    readable_xticklabels(ax)
    return figure2html(fig, close_fig=True)


def create_temporal_table(frame: pd.DataFrame, column: str, dt_column: str, period: str) -> str:
    r"""Create a HTML representation of a table with some statistics
    about the temporal value distribution.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.

    Returns:
        The HTML representation of the table.
    """
    if frame.shape[0] == 0:
        return "<span>&#9888;</span> No table is generated because the column is empty"
    frame = frame[[column, dt_column]].copy()
    dt_col = "__datetime__"
    frame[dt_col] = frame[dt_column].dt.to_period(period)
    frame_stats = (
        frame.groupby(dt_col)[column]
        .agg(
            [
                "count",
                "mean",
                "median",
                "min",
                "max",
                "std",
                ("q01", lambda x: x.quantile(0.01)),
                ("q05", lambda x: x.quantile(0.05)),
                ("q10", lambda x: x.quantile(0.1)),
                ("q25", lambda x: x.quantile(0.25)),
                ("q75", lambda x: x.quantile(0.75)),
                ("q90", lambda x: x.quantile(0.9)),
                ("q95", lambda x: x.quantile(0.95)),
                ("q99", lambda x: x.quantile(0.99)),
            ]
        )
        .sort_index()
    )

    rows = [create_temporal_table_row(row) for row in frame_stats.itertuples()]
    return Template(
        """
<details>
    <summary>[show statistics per temporal period]</summary>

    <p>The following table shows some statistics for each period of column {{column}}.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr>
                <th>column</th>
                <th>count</th>
                <th>mean</th>
                <th>std</th>
                <th>min</th>
                <th>quantile 1%</th>
                <th>quantile 5%</th>
                <th>quantile 10%</th>
                <th>quantile 25%</th>
                <th>median</th>
                <th>quantile 75%</th>
                <th>quantile 90%</th>
                <th>quantile 95%</th>
                <th>quantile 99%</th>
                <th>max</th>
            </tr>
        </thead>
        <tbody class="tbody table-group-divider">
            {{rows}}
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render({"rows": "\n".join(rows), "column": column, "period": period})


def create_temporal_table_row(row: pd.core.frame.Pandas) -> str:
    r"""Create the HTML code of a new table row.

    Args:
        row: Specifies a DataFrame row.

    Returns:
        The HTML code of a row.
    """
    return Template(
        """<tr>
    <th>{{datetime}}</th>
    <td {{num_style}}>{{count}}</td>
    <td {{num_style}}>{{mean}}</td>
    <td {{num_style}}>{{std}}</td>
    <td {{num_style}}>{{min}}</td>
    <td {{num_style}}>{{q01}}</td>
    <td {{num_style}}>{{q05}}</td>
    <td {{num_style}}>{{q10}}</td>
    <td {{num_style}}>{{q25}}</td>
    <td {{num_style}}>{{median}}</td>
    <td {{num_style}}>{{q75}}</td>
    <td {{num_style}}>{{q90}}</td>
    <td {{num_style}}>{{q95}}</td>
    <td {{num_style}}>{{q99}}</td>
    <td {{num_style}}>{{max}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "datetime": row.Index,
            "count": f"{row.count:,}",
            "mean": f"{row.mean:,.4f}",
            "median": f"{row.median:,.4f}",
            "min": f"{row.min:,.4f}",
            "max": f"{row.max:,.4f}",
            "std": f"{row.std:,.4f}",
            "q01": f"{row.q01:,.4f}",
            "q05": f"{row.q05:,.4f}",
            "q10": f"{row.q10:,.4f}",
            "q25": f"{row.q25:,.4f}",
            "q75": f"{row.q75:,.4f}",
            "q90": f"{row.q90:,.4f}",
            "q95": f"{row.q95:,.4f}",
            "q99": f"{row.q99:,.4f}",
        }
    )



r"""Contain the implementation of a section to analyze the temporal
distribution of null values for all columns."""

from __future__ import annotations

__all__ = ["TemporalNullValueSection"]

import logging
from typing import TYPE_CHECKING

import numpy as np
from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd
    from matplotlib.axes import Axes


logger = logging.getLogger(__name__)


class TemporalNullValueSection(BaseSection):
    r"""Implement a section to analyze the temporal distribution of null
    values for all columns.

    Args:
        frame: The DataFrame to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> import numpy as np
    >>> from flamme.section import TemporalNullValueSection
    >>> section = TemporalNullValueSection(
    ...     frame=pd.DataFrame(
    ...         {
    ...             "col1": np.array([1.2, 4.2, np.nan, 2.2]),
    ...             "col2": np.array([np.nan, 1, np.nan, 1]),
    ...             "datetime": pd.to_datetime(
    ...                 ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
    ...             ),
    ...         }
    ...     ),
    ...     columns=["col1", "col2"],
    ...     dt_column="datetime",
    ...     period="M",
    ... )
    >>> section
    TemporalNullValueSection(
      (columns): ('col1', 'col2')
      (dt_column): datetime
      (period): M
      (figsize): None
    )
    >>> section.get_statistics()
    {}

    ```
    """

    def __init__(
        self,
        frame: pd.DataFrame,
        columns: Sequence[str],
        dt_column: str,
        period: str,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        if dt_column not in frame:
            msg = (
                f"Datetime column {dt_column} is not in the DataFrame "
                f"(columns:{sorted(frame.columns)})"
            )
            raise ValueError(msg)

        self._frame = frame
        self._columns = tuple(columns)
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "columns": self._columns,
                    "dt_column": self._dt_column,
                    "period": self._period,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def frame(self) -> pd.DataFrame:
        r"""The DataFrame to analyze."""
        return self._frame

    @property
    def columns(self) -> tuple[str, ...]:
        r"""The columns to analyze."""
        return self._columns

    @property
    def dt_column(self) -> str:
        r"""The datetime column."""
        return self._dt_column

    @property
    def period(self) -> str:
        r"""The temporal period used to analyze the data."""
        return self._period

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            "Rendering the temporal distribution of null values for all columns "
            f"| datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "dt_column": self._dt_column,
                "figure": create_temporal_null_figure(
                    frame=self._frame,
                    columns=self._columns,
                    dt_column=self._dt_column,
                    period=self._period,
                    figsize=self._figsize,
                ),
                "table": create_temporal_null_table(
                    frame=self._frame,
                    columns=self._columns,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of null values in all columns.
The column <em>{{dt_column}}</em> is used as the temporal column.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_null_figure(
    frame: pd.DataFrame,
    columns: Sequence[str],
    dt_column: str,
    period: str,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Create a HTML representation of a figure with the temporal null
    value distribution.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze.
        dt_column: The datetime column used to analyze the
            temporal distribution.
        period: The temporal period e.g. monthly or daily.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The HTML representation of the figure.
    """
    if frame.shape[0] == 0:
        return ""

    num_nulls, total, labels = prepare_data(
        frame=frame, columns=columns, dt_column=dt_column, period=period
    )

    fig, ax = plt.subplots(figsize=figsize)
    plot_temporal_null_total(ax=ax, labels=labels, num_nulls=num_nulls, total=total)
    readable_xticklabels(ax, max_num_xticks=100)
    return figure2html(fig, close_fig=True)


def create_temporal_null_table(
    frame: pd.DataFrame, columns: Sequence[str], dt_column: str, period: str
) -> str:
    r"""Create a HTML representation of a table with the temporal
    distribution of null values.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze.
        dt_column: The datetime column used to analyze the
            temporal distribution.
        period: The temporal period e.g. monthly or daily.

    Returns:
        The HTML representation of the table.
    """
    if frame.shape[0] == 0:
        return ""
    num_nulls, totals, labels = prepare_data(
        frame=frame, columns=columns, dt_column=dt_column, period=period
    )
    rows = []
    for label, num_null, total in zip(labels, num_nulls, totals):
        rows.append(create_temporal_null_table_row(label=label, num_nulls=num_null, total=total))
    return Template(
        """
<details>
    <summary>[show statistics per temporal period]</summary>

    <p>The following table shows some statistics for each period.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr>
                <th>period</th>
                <th>number of null values</th>
                <th>number of non-null values</th>
                <th>total number of values</th>
                <th>percentage of null values</th>
                <th>percentage of non-null values</th>
            </tr>
        </thead>
        <tbody class="tbody table-group-divider">
            {{rows}}
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render({"rows": "\n".join(rows), "period": period})


def create_temporal_null_table_row(label: str, num_nulls: int, total: int) -> str:
    r"""Create the HTML code of a new table row.

    Args:
        label: The label of the row.
        num_nulls: The number of null values.
        total: The total number of values.

    Returns:
        The HTML code of a row.
    """
    num_non_nulls = total - num_nulls
    return Template(
        """<tr>
    <th>{{label}}</th>
    <td {{num_style}}>{{num_nulls}}</td>
    <td {{num_style}}>{{num_non_nulls}}</td>
    <td {{num_style}}>{{total}}</td>
    <td {{num_style}}>{{num_nulls_pct}}</td>
    <td {{num_style}}>{{num_non_nulls_pct}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "label": label,
            "num_nulls": f"{num_nulls:,}",
            "num_non_nulls": f"{num_non_nulls:,}",
            "total": f"{total:,}",
            "num_nulls_pct": f"{100 * num_nulls / total:.2f}%",
            "num_non_nulls_pct": f"{100 * num_non_nulls / total:.2f}%",
        }
    )


def plot_temporal_null_total(
    ax: Axes, num_nulls: np.ndarray, total: np.ndarray, labels: list
) -> None:
    color = "tab:blue"
    x = np.arange(len(labels))
    ax.set_ylabel("number of null/total values", color=color)
    ax.tick_params(axis="y", labelcolor=color)
    ax.bar(x=x, height=total, color="tab:cyan", alpha=0.5, label="total")
    ax.bar(x=x, height=num_nulls, color=color, alpha=0.8, label="null")
    ax.legend()

    ax2 = ax.twinx()
    color = "black"
    ax2.set_ylabel("percentage", color=color)
    ax2.tick_params(axis="y", labelcolor=color)
    ax2.plot(x, num_nulls / total, "o-", color=color)

    ax.set_xticks(x, labels=labels)
    ax.set_xlim(-0.5, len(labels) - 0.5)


def prepare_data(
    frame: pd.DataFrame,
    columns: Sequence[str],
    dt_column: str,
    period: str,
) -> tuple[np.ndarray, np.ndarray, list]:
    r"""Prepare the data to create the figure and table.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.

    Returns:
        A tuple with 3 values. The first value is a numpy NDArray
            that contains the number of null values per period. The
            second value is a numpy NDArray that contains the total
            number of values. The third value is a list that contains
            the label of each period.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.section.null_temp import prepare_data
    >>> num_nulls, total, labels = prepare_data(
    ...     frame=pd.DataFrame(
    ...         {
    ...             "col1": np.array([np.nan, 1, 0, 1]),
    ...             "col2": np.array([np.nan, 1, 0, np.nan]),
    ...             "datetime": pd.to_datetime(
    ...                 ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
    ...             ),
    ...         }
    ...     ),
    ...     columns=["col1", "col2"],
    ...     dt_column="datetime",
    ...     period="M",
    ... )
    >>> num_nulls
    array([2, 0, 0, 1])
    >>> total
    array([2, 2, 2, 2])
    >>> labels
    ['2020-01', '2020-02', '2020-03', '2020-04']

    ```
    """
    columns = list(columns)
    dt_col = "__datetime__"
    frame_na = frame[columns].isna().astype(float).copy()
    frame_na[dt_col] = frame[dt_column].dt.to_period(period)

    num_nulls = frame_na.groupby(dt_col)[columns].sum().sum(axis=1).sort_index()
    total = frame_na.groupby(dt_col)[columns].count().sum(axis=1).sort_index()
    labels = [str(dt) for dt in num_nulls.index]
    return num_nulls.to_numpy().astype(int), total.to_numpy().astype(int), labels



r"""Define some utility functions for testing."""

from __future__ import annotations

__all__ = [
    "clickhouse_connect_available",
]

import pytest

from flamme.utils.imports import is_clickhouse_connect_available

clickhouse_connect_available = pytest.mark.skipif(
    not is_clickhouse_connect_available(), reason="Requires clickhouse_connect"
)



r"""Contain the base class to implement a ``pandas.DataFrame``
transformer."""

from __future__ import annotations

__all__ = [
    "BaseDataFrameTransformer",
    "is_dataframe_transformer_config",
    "setup_dataframe_transformer",
]

import logging
from abc import ABC
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class BaseDataFrameTransformer(ABC, metaclass=AbstractFactory):
    r"""Define the base class to transform a ``pandas.DataFrame``.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.transformer.dataframe import ToNumeric
    >>> transformer = ToNumeric(columns=["col1", "col3"])
    >>> transformer
    ToNumericDataFrameTransformer(columns=('col1', 'col3'))
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame.dtypes
    col1     int64
    col2    object
    col3    object
    col4    object
    dtype: object
    >>> out = transformer.transform(frame)
    >>> out.dtypes
    col1     int64
    col2    object
    col3     int64
    col4    object
    dtype: object

    ```
    """

    def transform(self, frame: pd.DataFrame) -> pd.DataFrame:
        r"""Transform the data in the ``pandas.DataFrame``.

        Args:
            frame: The ``pandas.DataFrame`` to transform.

        Returns:
            The transformed DataFrame.

        Example usage:

        ```pycon
        >>> import pandas as pd
        >>> from flamme.transformer.dataframe import ToNumeric
        >>> transformer = ToNumeric(columns=["col1", "col3"])
        >>> frame = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> out = transformer.transform(frame)
        >>> out.dtypes
        col1     int64
        col2    object
        col3     int64
        col4    object
        dtype: object

        ```
        """


def is_dataframe_transformer_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseDataFrameTransformer``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration
            for a ``BaseDataFrameTransformer`` object.

    Example usage:

    ```pycon
    >>> from flamme.transformer.dataframe import is_dataframe_transformer_config
    >>> is_dataframe_transformer_config(
    ...     {"_target_": "flamme.transformer.dataframe.ToNumeric", "columns": ["col1", "col3"]}
    ... )
    True

    ```
    """
    return is_object_config(config, BaseDataFrameTransformer)


def setup_dataframe_transformer(
    transformer: BaseDataFrameTransformer | dict,
) -> BaseDataFrameTransformer:
    r"""Set up a ``pandas.DataFrame`` transformer.

    The transformer is instantiated from its configuration
    by using the ``BaseDataFrameTransformer`` factory function.

    Args:
        transformer: Specifies a ``pandas.DataFrame`` transformer or
            its configuration.

    Returns:
        An instantiated transformer.

    Example usage:

    ```pycon
    >>> from flamme.transformer.dataframe import setup_dataframe_transformer
    >>> transformer = setup_dataframe_transformer(
    ...     {"_target_": "flamme.transformer.dataframe.ToNumeric", "columns": ["col1", "col3"]}
    ... )
    >>> transformer
    ToNumericDataFrameTransformer(columns=('col1', 'col3'))

    ```
    """
    if isinstance(transformer, dict):
        logger.info("Initializing a DataFrame transformer from its configuration... ")
        transformer = BaseDataFrameTransformer.factory(**transformer)
    if not isinstance(transformer, BaseDataFrameTransformer):
        logger.warning(
            f"transformer is not a `BaseDataFrameTransformer` (received: {type(transformer)})"
        )
    return transformer



r"""Contain the base class to implement a reporter."""

from __future__ import annotations

__all__ = ["BaseReporter", "is_reporter_config", "setup_reporter"]

import logging
from abc import ABC

from objectory import AbstractFactory
from objectory.utils import is_object_config

logger = logging.getLogger(__name__)


class BaseReporter(ABC, metaclass=AbstractFactory):
    r"""Define the base class to compute a HTML report.

    Example usage:

    ```pycon
    >>> from flamme.analyzer import NullValueAnalyzer
    >>> from flamme.ingestor import ParquetIngestor
    >>> from flamme.transformer.dataframe import SequentialDataFrameTransformer
    >>> from flamme.reporter import Reporter
    >>> reporter = Reporter(
    ...     ingestor=ParquetIngestor("/path/to/data.parquet"),
    ...     transformer=SequentialDataFrameTransformer(transformers=[]),
    ...     analyzer=NullValueAnalyzer(),
    ...     report_path="/path/to/report.html",
    ... )
    >>> reporter
    Reporter(
      (ingestor): ParquetIngestor(path=/path/to/data.parquet)
      (transformer): SequentialDataFrameTransformer()
      (analyzer): NullValueAnalyzer(figsize=None)
      (report_path): /path/to/report.html
      (max_toc_depth): 6
    )
    >>> report = reporter.compute()  # doctest: +SKIP

    ```
    """

    def compute(self) -> None:
        r"""Generate a HTML report.

        Example usage:

        ```pycon
        >>> from flamme.analyzer import NullValueAnalyzer
        >>> from flamme.ingestor import ParquetIngestor
        >>> from flamme.transformer.dataframe import SequentialDataFrameTransformer
        >>> from flamme.reporter import Reporter
        >>> reporter = Reporter(
        ...     ingestor=ParquetIngestor("/path/to/data.parquet"),
        ...     transformer=SequentialDataFrameTransformer(transformers=[]),
        ...     analyzer=NullValueAnalyzer(figsize=None),
        ...     report_path="/path/to/report.html",
        ... )
        >>> report = reporter.compute()  # doctest: +SKIP

        ```
        """


def is_reporter_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseReporter``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseReporter`` object.

    Example usage:

    ```pycon
    >>> from flamme.reporter import is_reporter_config
    >>> is_reporter_config(
    ...     {
    ...         "_target_": "flamme.reporter.Reporter",
    ...         "ingestor": {
    ...             "_target_": "flamme.ingestor.CsvIngestor",
    ...             "path": "/path/to/data.csv",
    ...         },
    ...         "transformer": {
    ...             "_target_": "flamme.transformer.dataframe.ToNumeric",
    ...             "columns": ["col1", "col3"],
    ...         },
    ...         "analyzer": {"_target_": "flamme.analyzer.NullValueAnalyzer"},
    ...         "report_path": "/path/to/report.html",
    ...     }
    ... )
    True

    ```
    """
    return is_object_config(config, BaseReporter)


def setup_reporter(
    reporter: BaseReporter | dict,
) -> BaseReporter:
    r"""Set up a reporter.

    The reporter is instantiated from its configuration
    by using the ``BaseReporter`` factory function.

    Args:
        reporter: Specifies an reporter or its configuration.

    Returns:
        An instantiated reporter.

    Example usage:

    ```pycon
    >>> from flamme.reporter import setup_reporter
    >>> reporter = setup_reporter(
    ...     {
    ...         "_target_": "flamme.reporter.Reporter",
    ...         "ingestor": {
    ...             "_target_": "flamme.ingestor.CsvIngestor",
    ...             "path": "/path/to/data.csv",
    ...         },
    ...         "transformer": {
    ...             "_target_": "flamme.transformer.dataframe.ToNumeric",
    ...             "columns": ["col1", "col3"],
    ...         },
    ...         "analyzer": {"_target_": "flamme.analyzer.NullValueAnalyzer"},
    ...         "report_path": "/path/to/report.html",
    ...     }
    ... )
    >>> reporter
    Reporter(
      (ingestor): CsvIngestor(path=/path/to/data.csv)
      (transformer): ToNumericDataFrameTransformer(columns=('col1', 'col3'))
      (analyzer): NullValueAnalyzer(figsize=None)
      (report_path): /path/to/report.html
      (max_toc_depth): 6
    )

    ```
    """
    if isinstance(reporter, dict):
        logger.info("Initializing an reporter from its configuration... ")
        reporter = BaseReporter.factory(**reporter)
    if not isinstance(reporter, BaseReporter):
        logger.warning(f"reporter is not a `BaseReporter` (received: {type(reporter)})")
    return reporter



r"""Contain ``pandas.Series`` transformers."""

from __future__ import annotations

__all__ = [
    "BaseSeriesTransformer",
    "Sequential",
    "SequentialSeriesTransformer",
    "StripString",
    "StripStringSeriesTransformer",
    "ToDatetime",
    "ToDatetimeSeriesTransformer",
    "ToNumeric",
    "ToNumericSeriesTransformer",
    "is_series_transformer_config",
    "setup_series_transformer",
]

from flamme.transformer.series.base import (
    BaseSeriesTransformer,
    is_series_transformer_config,
    setup_series_transformer,
)
from flamme.transformer.series.datetime import ToDatetimeSeriesTransformer
from flamme.transformer.series.datetime import ToDatetimeSeriesTransformer as ToDatetime
from flamme.transformer.series.numeric import ToNumericSeriesTransformer
from flamme.transformer.series.numeric import ToNumericSeriesTransformer as ToNumeric
from flamme.transformer.series.sequential import SequentialSeriesTransformer
from flamme.transformer.series.sequential import (
    SequentialSeriesTransformer as Sequential,
)
from flamme.transformer.series.string import StripStringSeriesTransformer
from flamme.transformer.series.string import StripStringSeriesTransformer as StripString



r"""Implement an analyzer that generates a section to analyze the number
of null values."""

from __future__ import annotations

__all__ = ["NullValueAnalyzer"]

import logging
from typing import TYPE_CHECKING

import numpy as np

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import NullValueSection

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class NullValueAnalyzer(BaseAnalyzer):
    r"""Implement a null value analyzer.

    Args:
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import NullValueAnalyzer
    >>> analyzer = NullValueAnalyzer()
    >>> analyzer
    NullValueAnalyzer(figsize=None)
    >>> frame = pd.DataFrame(
    ...     {
    ...         "int": np.array([np.nan, 1, 0, 1]),
    ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...         "str": np.array(["A", "B", None, np.nan]),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(self, figsize: tuple[float, float] | None = None) -> None:
        self._figsize = figsize

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(figsize={self._figsize})"

    def analyze(self, frame: pd.DataFrame) -> NullValueSection:
        logger.info("Analyzing the null value distribution of all columns...")
        return NullValueSection(
            columns=list(frame.columns),
            null_count=frame.isna().sum().to_frame("count")["count"].to_numpy(),
            total_count=np.full((frame.shape[1],), frame.shape[0]),
            figsize=self._figsize,
        )



r"""Contain schema readers."""

from __future__ import annotations

__all__ = [
    "BaseSchemaReader",
    "ClickHouseSchemaReader",
    "ParquetSchemaReader",
    "SchemaReader",
    "is_schema_reader_config",
    "setup_schema_reader",
]

from flamme.schema.reader.base import (
    BaseSchemaReader,
    is_schema_reader_config,
    setup_schema_reader,
)
from flamme.schema.reader.clickhouse import ClickHouseSchemaReader
from flamme.schema.reader.parquet import ParquetSchemaReader
from flamme.schema.reader.vanilla import SchemaReader



r"""Implement an analyzer that generates a summary of the DataFrame."""

from __future__ import annotations

__all__ = ["DataFrameSummaryAnalyzer"]

import logging
from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import DataFrameSummarySection

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class DataFrameSummaryAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to show a summary of the DataFrame.

    Args:
        top: The number of most frequent values to show.
        sort: If ``True``, sort the columns by alphabetical order.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import DataFrameSummaryAnalyzer
    >>> analyzer = DataFrameSummaryAnalyzer()
    >>> analyzer
    DataFrameSummaryAnalyzer(top=5, sort=False)
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": np.array([0, 1, 0, 1]),
    ...         "col2": np.array([1, 0, 1, 0]),
    ...         "col3": np.array([1, 1, 1, 1]),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(self, top: int = 5, sort: bool = False) -> None:
        if top < 0:
            msg = f"Incorrect top value ({top}). top must be positive"
            raise ValueError(msg)
        self._top = top
        self._sort = bool(sort)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(top={self._top:,}, sort={self._sort})"

    def analyze(self, frame: pd.DataFrame) -> DataFrameSummarySection:
        logger.info("Analyzing the DataFrame...")
        if self._sort:
            frame = frame.reindex(sorted(frame.columns), axis=1)
        return DataFrameSummarySection(frame=frame, top=self._top)



r"""Contain utility functions to generate sections."""

from __future__ import annotations

__all__ = [
    "auto_yscale_continuous",
    "compute_statistics",
    "render_html_toc",
    "tags2id",
    "tags2title",
    "valid_h_tag",
]

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from scipy.stats import kurtosis, skew

from flamme.utils.array import nonnan

if TYPE_CHECKING:
    from collections.abc import Sequence

GO_TO_TOP = '<a href="#">Go to top</a>'


def tags2id(tags: Sequence[str]) -> str:
    r"""Convert a sequence of tags to a string that can be used as ID in
    a HTML file.

    Args:
        tags: The sequence of tags.

    Returns:
        The generated ID from the tags.
    """
    return "-".join(tags).replace(" ", "-").lower()


def tags2title(tags: Sequence[str]) -> str:
    r"""Convert a sequence of tags to a string that can be used as title.

    Args:
        tags: The sequence of tags.

    Returns:
        The generated title from the tags.
    """
    return " | ".join(tags[::-1])


def valid_h_tag(index: int) -> int:
    r"""Return a valid number of a h HTML tag.

    Args:
        index: The original value.

    Returns:
        A valid value.
    """
    return max(1, min(6, index))


def render_html_toc(
    number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
) -> str:
    r"""Return the HTML table of content (TOC) associated to the section.

    Args:
        number: The section number associated to the section.
        tags: The tags associated to the section.
        depth: The depth in the report.
        max_depth: The maximum depth to generate in the TOC.

    Returns:
        The HTML table of content associated to the section.
    """
    if depth >= max_depth:
        return ""
    tag = tags[-1] if tags else ""
    return f'<li><a href="#{tags2id(tags)}">{number} {tag}</a></li>'


def auto_yscale_continuous(array: np.ndarray, nbins: int | None = None) -> str:
    r"""Find a good scale for y-axis based on the data distribution.

    Args:
        array: The data to use to find the scale.
        nbins: The number of bins in the histogram.

    Returns:
        The scale for y-axis.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> from flamme.section.utils import auto_yscale_continuous
    >>> auto_yscale_continuous(np.arange(100))
    linear

    ```
    """
    if nbins is None:
        nbins = 100
    array = nonnan(array)
    counts = np.histogram(array, bins=nbins)[0]
    nonzero_count = [c for c in counts if c > 0]
    if len(nonzero_count) <= 2 or (max(nonzero_count) / max(min(nonzero_count), 1)) < 50:
        return "linear"
    if np.nanmin(array) <= 0.0:
        return "symlog"
    return "log"


def compute_statistics(data: pd.Series | np.ndarray) -> dict[str, float]:
    r"""Return several descriptive statistics for the input data.

    Args:
        data: The data to analyze.

    Returns:
        The descriptive statistics for the input data.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> from flamme.section.utils import compute_statistics
    >>> compute_statistics(np.arange(101))
    {'count': 101, 'num_nulls': 0, 'nunique': 101, 'mean': 50.0, 'std': 29.30...,
     'skewness': 0.0, 'kurtosis': -1.20..., 'min': 0.0, 'q001': 0.1, 'q01': 1.0,
     'q05': 5.0, 'q10': 10.0, 'q25': 25.0, 'median': 50.0, 'q75': 75.0, 'q90': 90.0,
     'q95': 95.0, 'q99': 99.0, 'q999': 99.9, 'max': 100.0, '>0': 100, '<0': 0, '=0': 1,
     'num_non_nulls': 101}

    ```
    """
    series = data if isinstance(data, pd.Series) else pd.Series(data)
    stats = {
        "count": int(series.shape[0]),
        "num_nulls": int(series.isna().sum()),
        "nunique": series.nunique(dropna=False),
        "mean": float("nan"),
        "std": float("nan"),
        "skewness": float("nan"),
        "kurtosis": float("nan"),
        "min": float("nan"),
        "q001": float("nan"),
        "q01": float("nan"),
        "q05": float("nan"),
        "q10": float("nan"),
        "q25": float("nan"),
        "median": float("nan"),
        "q75": float("nan"),
        "q90": float("nan"),
        "q95": float("nan"),
        "q99": float("nan"),
        "q999": float("nan"),
        "max": float("nan"),
        ">0": (series > 0).sum().item(),
        "<0": (series < 0).sum().item(),
        "=0": (series == 0).sum().item(),
    }
    stats["num_non_nulls"] = stats["count"] - stats["num_nulls"]
    if stats["num_non_nulls"] > 0:
        stats |= (
            series.dropna()
            .astype(float)
            .agg(
                {
                    "mean": "mean",
                    "median": "median",
                    "min": "min",
                    "max": "max",
                    "std": "std",
                    "q001": lambda x: x.quantile(0.001),
                    "q01": lambda x: x.quantile(0.01),
                    "q05": lambda x: x.quantile(0.05),
                    "q10": lambda x: x.quantile(0.1),
                    "q25": lambda x: x.quantile(0.25),
                    "q75": lambda x: x.quantile(0.75),
                    "q90": lambda x: x.quantile(0.9),
                    "q95": lambda x: x.quantile(0.95),
                    "q99": lambda x: x.quantile(0.99),
                    "q999": lambda x: x.quantile(0.999),
                }
            )
            .to_dict()
        )
        if stats["nunique"] > 1:
            array = data if isinstance(data, np.ndarray) else data.to_numpy(dtype=float)
            stats["skewness"] = float(skew(array, nan_policy="omit"))
            stats["kurtosis"] = float(kurtosis(array, nan_policy="omit"))
    return stats



r"""Contain ``pandas.DataFrame`` transformers to select columns in
DataFrames."""

from __future__ import annotations

__all__ = ["ColumnSelectionDataFrameTransformer"]

import logging
from typing import TYPE_CHECKING

from flamme.transformer.dataframe.base import BaseDataFrameTransformer

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd


logger = logging.getLogger(__name__)


class ColumnSelectionDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implement a ``pandas.DataFrame`` transformer to select a subset
    of columns.

    Args:
        columns: The columns to keep.
        ignore_missing: If ``False``, an exception is raised if a
            column is missing, otherwise a warning message is shown.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.transformer.dataframe import ColumnSelection
    >>> transformer = ColumnSelection(columns=["col1", "col2"])
    >>> transformer
    ColumnSelectionDataFrameTransformer(columns=['col1', 'col2'], ignore_missing=False)
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": ["2020-1-1", "2020-1-2", "2020-1-31", "2020-12-31", None],
    ...         "col2": [1, 2, 3, 4, 5],
    ...         "col3": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> out = transformer.transform(frame)
    >>> out
             col1  col2
    0    2020-1-1     1
    1    2020-1-2     2
    2   2020-1-31     3
    3  2020-12-31     4
    4        None     5

    ```
    """

    def __init__(self, columns: Sequence[str], ignore_missing: bool = False) -> None:
        self._columns = list(columns)
        self._ignore_missing = bool(ignore_missing)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(columns={self._columns}, "
            f"ignore_missing={self._ignore_missing})"
        )

    def transform(self, frame: pd.DataFrame) -> pd.DataFrame:
        columns = []
        for col in self._columns:
            if col not in frame:
                if self._ignore_missing:
                    logger.warning(f"Column `{col}` is not in the DataFrame")
                else:
                    msg = (
                        f"Column `{col}` is not in the DataFrame (columns:{sorted(frame.columns)})"
                    )
                    raise RuntimeError(msg)
            else:
                columns.append(col)
        logger.info(f"Selecting {len(columns):,} columns: {columns}")
        out = frame[columns].copy()
        logger.info(f"DataFrame shape after the column selection: {out.shape}")
        return out



r"""Contain a ``pandas.Series`` transformer to combine sequentially
multiple transformers."""

from __future__ import annotations

__all__ = ["SequentialSeriesTransformer"]

from typing import TYPE_CHECKING

from coola.utils import str_indent, str_sequence

from flamme.transformer.series import BaseSeriesTransformer, setup_series_transformer

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd


class SequentialSeriesTransformer(BaseSeriesTransformer):
    r"""Implement a ``pandas.Series`` transformer to apply sequentially
    several transformers.

    Args:
        transformers: The transformers or their
            configurations.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.transformer.series import Sequential, StripString, ToNumeric
    >>> transformer = Sequential([StripString(), ToNumeric()])
    >>> transformer
    SequentialSeriesTransformer(
      (0): StripStringSeriesTransformer()
      (1): ToNumericSeriesTransformer()
    )
    >>> series = pd.Series([" 1", "2 ", " 3 ", "4", "5"])
    >>> out = transformer.transform(series)
    >>> out
    0    1
    1    2
    2    3
    3    4
    4    5
    dtype: int64

    ```
    """

    def __init__(self, transformers: Sequence[BaseSeriesTransformer | dict]) -> None:
        self._transformers = tuple(
            setup_series_transformer(transformer) for transformer in transformers
        )

    def __repr__(self) -> str:
        args = ""
        if self._transformers:
            args = f"\n  {str_indent(str_sequence(self._transformers))}\n"
        return f"{self.__class__.__qualname__}({args})"

    def transform(self, series: pd.Series) -> pd.Series:
        for transformer in self._transformers:
            series = transformer.transform(series)
        return series



r"""Contain utility functions to manage data types."""

from __future__ import annotations

__all__ = [
    "frame_column_types",
    "find_date_columns_from_dtypes",
    "find_numeric_columns_from_dtypes",
    "get_dtypes_from_schema",
    "series_column_types",
]

import logging
from typing import TYPE_CHECKING

import pyarrow as pa

if TYPE_CHECKING:
    import pandas as pd


logger = logging.getLogger(__name__)


def frame_column_types(frame: pd.DataFrame) -> dict[str, set]:
    r"""Return the value types per column.

    Args:
        frame: The DataFrame to analyze.

    Returns:
        A dictionary with the value types for each column.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.utils.dtype import frame_column_types
    >>> frame = pd.DataFrame(
    ...     {
    ...         "int": np.array([np.nan, 1, 0, 1]),
    ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...     }
    ... )
    >>> coltypes = frame_column_types(frame)
    >>> coltypes
    {'int': {<class 'float'>}, 'float': {<class 'float'>}}

    ```
    """
    types = {}
    for col in frame:
        types[col] = series_column_types(frame[col])
    return types


def series_column_types(series: pd.Series) -> set[type]:
    r"""Return the value types in a ``pandas.Series``.

    Args:
        series: The DataFrame to analyze.

    Returns:
        A dictionary with the value types for each column.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.utils.dtype import series_column_types
    >>> coltypes = series_column_types(pd.Series([1.2, 4.2, np.nan, 2.2]))
    >>> coltypes
    {<class 'float'>}

    ```
    """
    return {type(x) for x in series.tolist()}


def find_numeric_columns_from_dtypes(dtypes: dict[str, pa.DataType]) -> list[str]:
    r"""Find the columns with a numeric type.

    Args:
        dtypes: The mapping of column names and data types.

    Returns:
        The columns with a numeric type.
    """
    columns = []
    for col, dtype in dtypes.items():
        if pa.types.is_decimal(dtype) or pa.types.is_floating(dtype) or pa.types.is_integer(dtype):
            columns.append(col)
    return columns


def find_date_columns_from_dtypes(dtypes: dict[str, pa.DataType]) -> list[str]:
    r"""Find the columns with a date type.

    Args:
        dtypes: The mapping of column names and data types.

    Returns:
        The columns with a date type.
    """
    columns = []
    for col, dtype in dtypes.items():
        if pa.types.is_date(dtype):
            columns.append(col)
    return columns


def get_dtypes_from_schema(schema: pa.Schema) -> dict[str, pa.DataType]:
    r"""Return the column data types from the schema.

    Args:
        schema: The table schema.

    Returns:
        The mapping of column names and data types.

    Example usage:

    ```pycon
    >>> import pyarrow
    >>> from flamme.utils.dtype import get_dtypes_from_schema
    >>> schema = pyarrow.schema([("number", pyarrow.int32()), ("string", pyarrow.string())])
    >>> dtypes = get_dtypes_from_schema(schema)
    >>> dtypes
    {'number': DataType(int32), 'string': DataType(string)}

    ```
    """
    return dict(zip(schema.names, schema.types))



r"""Implement analyzer that combine multiple analyzers."""

from __future__ import annotations

__all__ = ["MappingAnalyzer"]

from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer
from flamme.section import SectionDict

if TYPE_CHECKING:
    from collections.abc import Mapping

    import pandas as pd


class MappingAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer that combine multiple analyzers.

    Args:
        analyzers: The mappings to analyze.
            The key of each analyzer is used to organize the metrics
            and report.
        max_toc_depth: The maximum level to show in the
            table of content. Set this value to ``0`` to not show
            the table of content at the beginning of the section.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import (
    ...     FilteredAnalyzer,
    ...     NullValueAnalyzer,
    ...     DuplicatedRowAnalyzer,
    ...     MappingAnalyzer,
    ... )
    >>> analyzer = MappingAnalyzer(
    ...     {"null": NullValueAnalyzer(), "duplicate": DuplicatedRowAnalyzer()}
    ... )
    >>> analyzer
    MappingAnalyzer(
      (null): NullValueAnalyzer(figsize=None)
      (duplicate): DuplicatedRowAnalyzer(columns=None, figsize=None)
    )
    >>> frame = pd.DataFrame(
    ...     {
    ...         "int": np.array([np.nan, 1, 0, 1]),
    ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...         "str": np.array(["A", "B", None, np.nan]),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(
        self, analyzers: Mapping[str, BaseAnalyzer | dict], max_toc_depth: int = 0
    ) -> None:
        self._analyzers = {name: setup_analyzer(analyzer) for name, analyzer in analyzers.items()}
        self._max_toc_depth = max_toc_depth

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_mapping(self._analyzers))}\n)"

    @property
    def analyzers(self) -> dict[str, BaseAnalyzer]:
        return self._analyzers

    def analyze(self, frame: pd.DataFrame) -> SectionDict:
        return SectionDict(
            sections={name: analyzer.analyze(frame) for name, analyzer in self._analyzers.items()},
            max_toc_depth=self._max_toc_depth,
        )

    def add_analyzer(self, key: str, analyzer: BaseAnalyzer, replace_ok: bool = False) -> None:
        r"""Add an analyzer to the current analyzer.

        Args:
            key: The key of the analyzer.
            analyzer: The analyzer to add.
            replace_ok: If ``False``, ``KeyError`` is raised if an
                analyzer with the same key exists. If ``True``,
                the new analyzer will replace the existing analyzer.

        Raises:
            KeyError: if an  analyzer with the same key exists.

        Example usage:

        ```pycon
        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import MappingAnalyzer, NullValueAnalyzer, DuplicatedRowAnalyzer
        >>> analyzer = MappingAnalyzer({"null": NullValueAnalyzer()})
        >>> analyzer.add_analyzer("duplicate", DuplicatedRowAnalyzer())
        >>> analyzer
        MappingAnalyzer(
          (null): NullValueAnalyzer(figsize=None)
          (duplicate): DuplicatedRowAnalyzer(columns=None, figsize=None)
        )
        >>> frame = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(frame)

        ```
        """
        if key in self._analyzers and not replace_ok:
            msg = (
                f"`{key}` is already used to register an analyzer. "
                "Use `replace_ok=True` to replace the analyzer"
            )
            raise KeyError(msg)
        self._analyzers[key] = analyzer



r"""Contain the implementation of a simple reporter."""

from __future__ import annotations

__all__ = ["Reporter"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer
from flamme.ingestor.base import BaseIngestor, setup_ingestor
from flamme.reporter.base import BaseReporter
from flamme.reporter.utils import create_html_report
from flamme.transformer.dataframe.base import (
    BaseDataFrameTransformer,
    setup_dataframe_transformer,
)
from flamme.utils.io import save_text
from flamme.utils.path import sanitize_path

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


class Reporter(BaseReporter):
    r"""Implement a simple reporter.

    Args:
        ingestor: The ingestor or its configuration.
        transformer: Specifies a ``pandas.DataFrame`` transformer or
            its configuration.
        ingestor: The analyzer or its configuration.
        report_path: The path where to save the HTML report.
        max_toc_depth: The maximum level to show in the
            table of content.

    Example usage:

    ```pycon
    >>> from flamme.analyzer import NullValueAnalyzer
    >>> from flamme.ingestor import ParquetIngestor
    >>> from flamme.transformer.dataframe import SequentialDataFrameTransformer
    >>> from flamme.reporter import Reporter
    >>> reporter = Reporter(
    ...     ingestor=ParquetIngestor("/path/to/data.parquet"),
    ...     transformer=SequentialDataFrameTransformer(transformers=[]),
    ...     analyzer=NullValueAnalyzer(),
    ...     report_path="/path/to/report.html",
    ... )
    >>> report = reporter.compute()  # doctest: +SKIP

    ```
    """

    def __init__(
        self,
        ingestor: BaseIngestor | dict,
        transformer: BaseDataFrameTransformer | dict,
        analyzer: BaseAnalyzer | dict,
        report_path: Path | str,
        max_toc_depth: int = 6,
    ) -> None:
        self._ingestor = setup_ingestor(ingestor)
        logger.info(f"ingestor:\n{ingestor}")
        self._transformer = setup_dataframe_transformer(transformer)
        logger.info(f"transformer:\n{transformer}")
        self._analyzer = setup_analyzer(analyzer)
        logger.info(f"analyzer:\n{analyzer}")
        self._report_path = sanitize_path(report_path)
        self._max_toc_depth = int(max_toc_depth)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "ingestor": self._ingestor,
                    "transformer": self._transformer,
                    "analyzer": self._analyzer,
                    "report_path": self._report_path,
                    "max_toc_depth": self._max_toc_depth,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def compute(self) -> None:
        logger.info("Ingesting the DataFrame...")
        frame = self._ingestor.ingest()
        logger.info(f"Transforming the DataFrame {frame.shape}...")
        frame = self._transformer.transform(frame)
        logger.info(f"Analyzing the DataFrame {frame.shape}...")
        section = self._analyzer.analyze(frame)
        logger.info("Creating the HTML report...")
        report = create_html_report(
            toc=section.render_html_toc(max_depth=self._max_toc_depth),
            body=section.render_html_body(),
        )
        logger.info(f"Saving HTML report at {self._report_path}...")
        save_text(report, self._report_path)



r"""Contain ``pandas.DataFrame`` transformers that combines multiple
``pandas.Series`` transformers."""

from __future__ import annotations

__all__ = ["ColumnDataFrameTransformer"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping
from tqdm import tqdm

from flamme.transformer.dataframe.base import BaseDataFrameTransformer

if TYPE_CHECKING:

    import pandas as pd

    from flamme.transformer.series.base import BaseSeriesTransformer

logger = logging.getLogger(__name__)


class ColumnDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implement a ``pandas.DataFrame`` transformer that applies
    ``pandas.Series`` transformers on some columns.

    Args:
        columns: The ``pandas.Series`` transformers.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.transformer.dataframe import Column
    >>> from flamme.transformer.series import ToNumeric, StripString
    >>> transformer = Column({"col2": ToNumeric(), "col3": StripString()})
    >>> transformer
    ColumnDataFrameTransformer(
      (col2): ToNumericSeriesTransformer()
      (col3): StripStringSeriesTransformer()
      (ignore_missing): False
    )
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": [" a", "b ", " c ", "  d  ", "e"],
    ...     }
    ... )
    >>> out = transformer.transform(frame)
    >>> out
       col1  col2 col3
    0     1     1    a
    1     2     2    b
    2     3     3    c
    3     4     4    d
    4     5     5    e

    ```
    """

    def __init__(
        self, columns: dict[str, BaseSeriesTransformer | dict], ignore_missing: bool = False
    ) -> None:
        self._columns = columns
        self._ignore_missing = bool(ignore_missing)

    def __repr__(self) -> str:
        columns = self._columns | {"ignore_missing": self._ignore_missing}
        args = f"\n  {str_indent(str_mapping(columns))}\n"
        return f"{self.__class__.__qualname__}({args})"

    @property
    def transformers(self) -> dict[str, BaseSeriesTransformer]:
        return self._columns

    def transform(self, frame: pd.DataFrame) -> pd.DataFrame:
        for col, transformer in tqdm(self._columns.items(), desc="Transforming columns"):
            if col not in frame:
                if self._ignore_missing:
                    logger.warning(
                        f"Skipping transformation for column {col} because the column is missing"
                    )
                else:
                    msg = f"Column {col} is not in the DataFrame (columns:{sorted(frame.columns)})"
                    raise RuntimeError(msg)
            else:
                logger.info(f"Transforming column `{col}`...")
                frame[col] = transformer.transform(frame[col])
        return frame

    def add_transformer(
        self, column: str, transformer: BaseSeriesTransformer, replace_ok: bool = False
    ) -> None:
        r"""Add a transformer to the current analyzer.

        Args:
            column: The column of the transformer.
            transformer: The transformer to add.
            replace_ok: If ``False``, ``KeyError`` is raised if a
                transformer with the same key exists. If ``True``,
                the new transformer will replace the existing
                transformer.

        Raises:
            KeyError: if a transformer with the same key exists.

        Example usage:

        ```pycon
        >>> import pandas as pd
        >>> from flamme.transformer.dataframe import Column
        >>> from flamme.transformer.series import ToNumeric, StripString
        >>> transformer = Column({"col2": ToNumeric(), "col3": StripString()})
        >>> transformer
        ColumnDataFrameTransformer(
          (col2): ToNumericSeriesTransformer()
          (col3): StripStringSeriesTransformer()
          (ignore_missing): False
        )
        >>> transformer.add_transformer("col1", ToNumeric())
        >>> transformer
        ColumnDataFrameTransformer(
          (col2): ToNumericSeriesTransformer()
          (col3): StripStringSeriesTransformer()
          (col1): ToNumericSeriesTransformer()
          (ignore_missing): False
        )

        ```
        """
        if column in self._columns and not replace_ok:
            msg = (
                f"`{column}` is already used to register a transformer. "
                "Use `replace_ok=True` to replace the transformer"
            )
            raise KeyError(msg)
        self._columns[column] = transformer



r"""Contain utility functions to analyze data with null values."""

from __future__ import annotations

__all__ = ["compute_null_per_col"]

import numpy as np
import pandas as pd


def compute_null_per_col(frame: pd.DataFrame) -> pd.DataFrame:
    r"""Return the number and percentage of null values per column.

    Args:
        frame: The DataFrame to analyze.

    Returns:
        A DataFrame with the number and percentage of null values per
            column.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.utils.null import compute_null_per_col
    >>> frame = compute_null_per_col(
    ...     pd.DataFrame(
    ...         {
    ...             "int": np.array([np.nan, 1, 0, 1]),
    ...             "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...             "str": np.array(["A", "B", None, np.nan]),
    ...         }
    ...     )
    ... )
    >>> frame
      column  null  total  null_pct
    0    int     1      4      0.25
    1  float     1      4      0.25
    2    str     2      4      0.50

    ```
    """
    null_count = frame.isna().sum().to_frame("count")["count"].to_numpy().astype(int)
    total_count = np.full((frame.shape[1],), frame.shape[0]).astype(int)
    with np.errstate(invalid="ignore"):
        null_pct = null_count.astype(float) / total_count.astype(float)
    return pd.DataFrame(
        {
            "column": list(frame.columns),
            "null": null_count,
            "total": total_count,
            "null_pct": null_pct,
        }
    )



r"""Contain ``pandas.DataFrame`` transformers to transform columns with
numeric values."""

from __future__ import annotations

__all__ = ["ToNumericDataFrameTransformer"]

import logging
from typing import TYPE_CHECKING, Any

import pandas as pd
from tqdm import tqdm

from flamme.transformer.dataframe.base import BaseDataFrameTransformer
from flamme.utils.dtype import find_numeric_columns_from_dtypes, get_dtypes_from_schema

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pyarrow as pa

logger = logging.getLogger(__name__)


class ToNumericDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implement a transformer to convert some columns to numeric type.

    Args:
        columns: The columns to convert.
        **kwargs: The keyword arguments for ``pandas.to_numeric``.

    Example usage:

    ```pycon

    >>> import pandas as pd
    >>> from flamme.transformer.dataframe import ToNumeric
    >>> transformer = ToNumeric(columns=["col1", "col3"])
    >>> transformer
    ToNumericDataFrameTransformer(columns=('col1', 'col3'))
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame.dtypes
    col1     int64
    col2    object
    col3    object
    col4    object
    dtype: object
    >>> out = transformer.transform(frame)
    >>> out.dtypes
    col1     int64
    col2    object
    col3     int64
    col4    object
    dtype: object

    ```
    """

    def __init__(self, columns: Sequence[str], **kwargs: Any) -> None:
        self._columns = tuple(columns)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = ", ".join([f"{key}={value}" for key, value in self._kwargs.items()])
        if args:
            args = ", " + args
        return f"{self.__class__.__qualname__}(columns={self._columns}{args})"

    def transform(self, frame: pd.DataFrame) -> pd.DataFrame:
        for col in tqdm(self._columns, desc="Converting to numeric"):
            frame[col] = pd.to_numeric(frame[col], **self._kwargs)
        return frame

    @classmethod
    def from_schema(cls, schema: pa.Schema, **kwargs: Any) -> ToNumericDataFrameTransformer:
        r"""Instantiate a ``ToNumericDataFrameTransformer`` where the
        columns are automatically selected from the schema.

        Args:
            schema: The DataFrame schema.
            **kwargs: The keyword arguments for ``pandas.to_numeric``.

        Returns:
            An instantiated ``ToNumericDataFrameTransformer``.

        Example usage:

        ```pycon

        >>> import pandas as pd
        >>> import pyarrow as pa
        >>> from flamme.transformer.dataframe import ToNumeric
        >>> transformer = ToNumeric.from_schema(
        ...     pa.schema([("col1", pa.int64()), ("col2", pa.string()), ("col3", pa.int64())])
        ... )
        >>> transformer
        ToNumericDataFrameTransformer(columns=('col1', 'col3'))
        >>> frame = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> frame.dtypes
        col1     int64
        col2    object
        col3    object
        col4    object
        dtype: object
        >>> out = transformer.transform(frame)
        >>> out.dtypes
        col1     int64
        col2    object
        col3     int64
        col4    object
        dtype: object

        ```
        """
        dtypes = get_dtypes_from_schema(schema)
        columns = sorted(find_numeric_columns_from_dtypes(dtypes))
        logger.info(f"found {len(columns):,} numeric columns: {columns}")
        return cls(columns=columns, **kwargs)



r"""Contain some clickhouse utility functions."""

from __future__ import annotations

__all__ = ["get_table_schema", "get_dtypes"]


from typing import TYPE_CHECKING

import pyarrow as pa

from flamme.utils.dtype import get_dtypes_from_schema
from flamme.utils.path import sanitize_path

if TYPE_CHECKING:
    from pathlib import Path


def get_table_schema(path: Path | str) -> pa.Schema:
    r"""Return the table schema.

    Args:
        path: The path to the parquet file.

    Returns:
        The table schema.
    """
    return pa.parquet.read_schema(sanitize_path(path))


def get_dtypes(path: Path | str) -> dict[str, pa.DataType]:
    r"""Return the column data types from the schema.

    Args:
        path: The path to the schema.

    Returns:
        The mapping of column names and data types.
    """
    return get_dtypes_from_schema(schema=get_table_schema(path))



r"""Contain some clickhouse utility functions."""

from __future__ import annotations

__all__ = ["get_table_schema"]


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from flamme.utils.imports import is_clickhouse_connect_available

    if is_clickhouse_connect_available():  # pragma: no cover
        from clickhouse_connect.driver.client import Client
    import pyarrow as pa


def get_table_schema(client: Client, table: str) -> pa.Schema:
    r"""Return the table schema.

    Args:
        client: The clickhouse client.
        table: The table.

    Returns:
        The table schema.
    """
    return client.query_arrow(query=f"select top 1 * from {table}").schema  # noqa: S608



r"""Contain ``pandas.DataFrame`` transformers to transform columns with
null values."""

from __future__ import annotations

__all__ = ["NullColumnDataFrameTransformer"]

import logging
from typing import TYPE_CHECKING

from flamme.transformer.dataframe.base import BaseDataFrameTransformer
from flamme.utils.null import compute_null_per_col

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class NullColumnDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implement a ``pandas.DataFrame`` transformer to remove the
    columns that have too many null values.

    Args:
        threshold (float): The maximum percentage of null
            values to keep columns. If the proportion of null vallues
            is greater or equal to this threshold value, the column
            is removed. If set to ``1.0``, it removes all the columns
            that have only null values.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.dataframe import NullColumn
        >>> transformer = NullColumn(threshold=1.0)
        >>> transformer
        NullColumnDataFrameTransformer(threshold=1.0)
        >>> frame = pd.DataFrame(
        ...     {
        ...         "col1": ["2020-1-1", "2020-1-2", "2020-1-31", "2020-12-31", None],
        ...         "col2": [1, None, 3, None, 5],
        ...         "col3": [None, None, None, None, None],
        ...     }
        ... )
        >>> out = transformer.transform(frame)
        >>> out
                 col1  col2
        0    2020-1-1   1.0
        1    2020-1-2   NaN
        2   2020-1-31   3.0
        3  2020-12-31   NaN
        4        None   5.0
    """

    def __init__(self, threshold: float) -> None:
        self._threshold = float(threshold)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(threshold={self._threshold})"

    def transform(self, frame: pd.DataFrame) -> pd.DataFrame:
        if frame.shape[0] == 0:
            return frame
        num_orig_cols = len(frame.columns)
        frame_null = compute_null_per_col(frame)
        columns = frame_null[frame_null["null_pct"] < self._threshold]["column"].tolist()
        logger.info(
            f"Removing {num_orig_cols - len(columns):,} columns because they have too "
            f"many null values (threshold={self._threshold})..."
        )
        return frame[columns].copy()



r"""Implement an analyzer that generates a section to analyze number of
duplicated rows."""

from __future__ import annotations

__all__ = ["DuplicatedRowAnalyzer"]

import logging
from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import DuplicatedRowSection

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd


logger = logging.getLogger(__name__)


class DuplicatedRowAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to show the number of duplicated rows.

    Args:
        columns: The columns used to compute the duplicated
            rows. ``None`` means all the columns.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import DuplicatedRowAnalyzer
    >>> analyzer = DuplicatedRowAnalyzer()
    >>> analyzer
    DuplicatedRowAnalyzer(columns=None, figsize=None)
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": np.array([0, 1, 0, 1]),
    ...         "col2": np.array([1, 0, 1, 0]),
    ...         "col3": np.array([1, 1, 1, 1]),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None = None,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._columns = columns
        self._figsize = figsize

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(columns={self._columns}, figsize={self._figsize})"

    def analyze(self, frame: pd.DataFrame) -> DuplicatedRowSection:
        logger.info(f"Analyzing the duplicated rows section using the columns: {self._columns}")
        return DuplicatedRowSection(frame=frame, columns=self._columns, figsize=self._figsize)



r"""Implement a section that generates a table of content."""

from __future__ import annotations

__all__ = ["TableOfContentSection"]

from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping

from flamme.section.base import BaseSection

if TYPE_CHECKING:
    from collections.abc import Sequence


class TableOfContentSection(BaseSection):
    r"""Implement a wrapper section that generates a table of content
    before the section.

    Args:
        section: The section.
        max_toc_depth: The maximum level to show in the
            table of content.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> import numpy as np
    >>> from flamme.section import TableOfContentSection, DuplicatedRowSection
    >>> section = TableOfContentSection(
    ...     DuplicatedRowSection(
    ...         frame=pd.DataFrame(
    ...             {
    ...                 "col1": np.array([1.2, 4.2, 4.2, 2.2]),
    ...                 "col2": np.array([1, 1, 1, 1]),
    ...                 "col3": np.array([1, 2, 2, 2]),
    ...             }
    ...         )
    ...     )
    ... )
    >>> section
    TableOfContentSection(
      (section): DuplicatedRowSection(
          (columns): None
          (figsize): None
        )
      (max_toc_depth): 1
    )
    >>> section.get_statistics()
    {'num_rows': 4, 'num_unique_rows': 3}

    ```
    """

    def __init__(self, section: BaseSection, max_toc_depth: int = 1) -> None:
        self._section = section
        self._max_toc_depth = int(max_toc_depth)

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping({"section": self._section, "max_toc_depth": self._max_toc_depth})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def get_statistics(self) -> dict:
        return self._section.get_statistics()

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        toc = self._section.render_html_toc(
            number=number, tags=tags, depth=0, max_depth=self._max_toc_depth
        )
        body = self._section.render_html_body(number=number, tags=tags, depth=depth)
        return f"{toc}\n\n{body}"

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return self._section.render_html_toc(
            number=number, tags=tags, depth=depth, max_depth=max_depth
        )
